package be.iminds.ilabt.jfed.ssh_terminal_tool;

import be.iminds.ilabt.jfed.gui_preferences.HLPreferenceKey;
import be.iminds.ilabt.jfed.highlevel.util.SshCommandBuilder;
import be.iminds.ilabt.jfed.highlevel.util.SshCommandBuilderFactory;
import be.iminds.ilabt.jfed.lowlevel.user.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.user.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnection;
import be.iminds.ilabt.jfed.lowlevel.ssh_key_info.PuTTYFilesKeyInfo;
import be.iminds.ilabt.jfed.lowlevel.ssh_key_info.SshKeyInfoFactory;
import be.iminds.ilabt.jfed.preferences.CorePreferenceKey;
import be.iminds.ilabt.jfed.gui_preferences.JFedHLPreferences;
import be.iminds.ilabt.jfed.preferences.ProxyPreferencesManager;
import be.iminds.ilabt.jfed.ssh_terminal_tool.putty.PuttyHelper;
import be.iminds.ilabt.jfed.ssh_terminal_tool.known_hosts.OpenSshKnownHostsFile;
import be.iminds.ilabt.jfed.ssh_terminal_tool.unix.SshAgentHelper;
import be.iminds.ilabt.jfed.ui.javafx.GlyphUtils;
import be.iminds.ilabt.jfed.util.common.OSDetector;
import be.iminds.ilabt.jfed.util.common.Pair;
import be.iminds.ilabt.jfed.util.tmp_file_helpers.TmpContentFile;
import be.iminds.ilabt.jfed.util.tmp_file_helpers.TmpFile;
import javafx.scene.control.*;
import javafx.scene.layout.VBox;
import javafx.scene.paint.Color;
import org.controlsfx.glyphfont.FontAwesome;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.inject.Inject;
import javax.inject.Singleton;
import javax.script.ScriptEngine;
import javax.script.ScriptEngineManager;
import javax.script.ScriptException;
import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Objects;
import java.util.regex.Pattern;

@Singleton
public class ExternalSshTerminal {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(ExternalSshTerminal.class);

    private final GeniUserProvider geniUserProvider;
    private final JFedHLPreferences jFedPreferences;
    private final ProxyPreferencesManager proxyPreferencesManager;
    private final PuttyHelper puttyHelper;
    private final OpenSshKnownHostsFile openSshKnownHostsFile;
    private final SshCommandBuilderFactory sshCommandBuilderFactory;

    @Inject
    public ExternalSshTerminal(GeniUserProvider geniUserProvider,
                               ProxyPreferencesManager proxyPreferencesManager,
                               JFedHLPreferences jFedPreferences,
                               PuttyHelper puttyHelper,
                               OpenSshKnownHostsFile openSshKnownHostsFile,
                               SshCommandBuilderFactory sshCommandBuilderFactory) {
        this.geniUserProvider = geniUserProvider;
        this.proxyPreferencesManager = proxyPreferencesManager;
        this.jFedPreferences = jFedPreferences;
        this.puttyHelper = puttyHelper;
        this.openSshKnownHostsFile = openSshKnownHostsFile;
        this.sshCommandBuilderFactory = sshCommandBuilderFactory;

        //not the best way, but currently the easiest way to activate more detailed logging only for ExternalSshTerminal
        boolean terminalDebug = jFedPreferences.getBoolean(HLPreferenceKey.PREF_TERMINAL_DEBUG, true);
        //TODO: re-enable this, by actually configuring the logging differently
    }

    public boolean isConfigured() {
        LOG.debug("ExternalSshTerminal.isConfigured");
        if (OSDetector.os == OSDetector.OS.WIN) {
            File puttyDir = jFedPreferences.getFile(HLPreferenceKey.PREF_PUTTY_DIRECTORY);
            LOG.debug("ExternalSshTerminal.isConfigured WIN puttyDir=" + puttyDir);
            if (puttyDir == null) return false;
            boolean valid = JFedHLPreferences.isValidPuttyDir(puttyDir);
            LOG.debug("ExternalSshTerminal.isConfigured WIN puttyDir=" + puttyDir + " valid=" + valid);
            return valid;
        }

        if (OSDetector.os == OSDetector.OS.UNIX) {
            return jFedPreferences.getString(HLPreferenceKey.PREF_COMMAND_UNIX_TERMINAL) != null;
        }

        if (OSDetector.os == OSDetector.OS.MAC) {
            return jFedPreferences.getString(HLPreferenceKey.PREF_COMMAND_UNIX_TERMINAL) != null;
        }

        return true;
    }

    public String getCliSshCommand(@Nonnull String targetServerUserName, @Nonnull String hostName, int port,
                                   @Nullable JFedConnection.ProxyInfo customProxyInfo) {
        SshCommandBuilder builder = prepareSshCommand(
                targetServerUserName,
                hostName,
                port,
                customProxyInfo,
                DoubleProxyHandling.DIALOG);

        return builder.build(true).getKey();
    }

    public void launch(@Nonnull String targetServerUserName, @Nonnull String hostName, int port,
                       @Nullable JFedConnection.ProxyInfo customProxyInfo) {
        SshCommandBuilder sshCommandBuilder = prepareSshCommand(
                targetServerUserName,
                hostName,
                port,
                customProxyInfo,
                DoubleProxyHandling.DIALOG);

        if (sshCommandBuilder.getProxyInfo() == null)
            LOG.info("ExternalSshTerminal: No proxy will be used.");
        else
            LOG.info("ExternalSshTerminal: Proxy will be used: " + sshCommandBuilder.getProxyInfo());

        if (OSDetector.os == OSDetector.OS.WIN) {
            launchPutty(targetServerUserName,
                    hostName,
                    port,
                    sshCommandBuilder.getSshKeyInfos(),
                    sshCommandBuilder.getProxyInfo());
            return;
        }

        if (OSDetector.os == OSDetector.OS.UNIX) {
            launchUnixOrMacTerminal(false,
                    sshCommandBuilder);
            return;
        }

        if (OSDetector.os == OSDetector.OS.MAC) {
            launchUnixOrMacTerminal(
                    true,
                    sshCommandBuilder);
            return;
        }

        throw new RuntimeException("No terminal known for OS \"" + OSDetector.os + "\"");
    }

    public void launchPutty(final String userName, final String hostName, final int port, final List<? extends PuTTYFilesKeyInfo> sshKeyInfos,
                            final JFedConnection.ProxyInfo proxyInfo) {
        PuttyHelper.PuttyRunParameters parameters = new PuttyHelper.PuttyRunParameters();
        parameters.setTarget(hostName, port, sshKeyInfos, userName);
        parameters.setProxy(proxyInfo);
        parameters.setEnableX11Forward(jFedPreferences.getBoolean(HLPreferenceKey.PREF_PUTTY_X11_FORWARD, true));
        parameters.setEnableAgentForward(jFedPreferences.getBoolean(HLPreferenceKey.PREF_PUTTY_AGENT_FORWARD, true));
        puttyHelper.runPutty(parameters);
    }


    public void launchUnixOrMacTerminal(boolean isMac,
                                        SshCommandBuilder sshCommandBuilder) {
        if (isMac && jFedPreferences.getString(HLPreferenceKey.PREF_COMMAND_UNIX_TERMINAL) == null) {
            String com = "tell application \"Terminal\" to do script \"%\"";

            LOG.info("Falling back to auto-detected terminal command: " + com);
            jFedPreferences.setString(HLPreferenceKey.PREF_COMMAND_UNIX_TERMINAL, com);
        }

        //for SSH proxy to work, we need to store the private key unencrypted.
        final List<TmpFile> tmpFiles = new ArrayList<>();

        String hostName = sshCommandBuilder.getHostName();
        Pair<String, List<TmpFile>> sshComPair = sshCommandBuilder.build(false);
        String innerSshCommand = sshComPair.getKey();
        tmpFiles.addAll(sshComPair.getValue());

        //set -v shows commands executed
        //set +v stops showing commands executed
        // (+x and -x hide and show expanded commands)
        TmpContentFile tmpCommandFile = null;
        if (sshCommandBuilder.isUseSshAgent())
            try {
                tmpCommandFile =
                        new TmpContentFile("openSsh", "sh",
                                "#!/bin/bash \n\n" +
                                        "set -v\n" +
                                        SshAgentHelper.getSshAgentEnvironment() + "\n" +
                                        innerSshCommand + "\n" +
                                        "/bin/bash");
            } catch (SshAgentHelper.SshAgentHelperException e) {
                LOG.error("Problem launching SSH terminal: Failed to add SSH agent info to tmpCommandFile. Will skip using SSH agent! (this goes against user preference!)", e);
            }
        if (tmpCommandFile == null)
            tmpCommandFile =
                    new TmpContentFile("openSsh", "sh", "#!/bin/bash \n\nset -v\n" + innerSshCommand + "\n/bin/bash");
        tmpFiles.add(tmpCommandFile);
        String tmpCommandFilename = tmpCommandFile.getFilename();
        //        String quotedTmpCommandFilename = "'"+tmpCommandFileHelper.getFilename()+"'";

        final String callScriptCommand = "bash " + tmpCommandFilename;

        String terminalCommand = jFedPreferences.getString(HLPreferenceKey.PREF_COMMAND_UNIX_TERMINAL);
        if (terminalCommand == null)
            throw new RuntimeException("No termincal command specified in jFed config. Cannot launch terminal.");

        final List<String> launchTerminalCommandList = new ArrayList<>();
        if (isMac) {
            //commandList not used
            if (terminalCommand.contains("%")) {
                terminalCommand = terminalCommand.replaceFirst("%", callScriptCommand);
            } else
                LOG.error("command does not contain required % \"" + terminalCommand + "\"");
        } else {
            //unix
            if (!terminalCommand.contains("%")) {
                //We do not know the exact command, so we will assume it is '-e' which is what xterm and konsole use
                //mustQuote is a hack to make gnome-terminal work (it needs quotes around the command)
                boolean mustQuote = false; //false for xterm and konsole, true for gnome-terminal
                if (innerSshCommand.contains("gnome-terminal"))
                    mustQuote = true;

                if (mustQuote)
                    terminalCommand += " -e '%'";
                else
                    terminalCommand += " -e %";
            }

            if (terminalCommand.contains("'%'") || terminalCommand.contains("\"%\"")) {
                terminalCommand = terminalCommand.replaceFirst(Pattern.quote("'%'"), "%").replaceFirst(Pattern.quote("\"%\""), "%");

                String[] finalCommandSplit = terminalCommand.split(" ");
                for (String aFinalCommandSplit : finalCommandSplit)
                    if (Objects.equals(aFinalCommandSplit, "%"))
                        launchTerminalCommandList.add(callScriptCommand);
                    else
                        launchTerminalCommandList.add(aFinalCommandSplit);
            } else {
                terminalCommand = terminalCommand.replaceFirst(Pattern.quote("%"), callScriptCommand);
                String[] finalCommandSplit = terminalCommand.split(" ");
                Collections.addAll(launchTerminalCommandList, finalCommandSplit);
            }
        }
        final String finalCommand = terminalCommand;

        LOG.debug("launchTerminal: useSshAgent=" + sshCommandBuilder.isUseSshAgent() + " proxyInfo=" + sshCommandBuilder.getProxyInfo());
        LOG.info("launchTerminal: commands={}", tmpCommandFile.getContent());
        if (isMac) {
            LOG.debug("launchTerminal: callScriptCommand: " + callScriptCommand);
        } else {
            LOG.debug("Executing command to launch terminal: {}", terminalCommand);
            LOG.debug("   internalCommand: {}", innerSshCommand);

            LOG.info("launchTerminal: commandList={}", launchTerminalCommandList);

            //example commands:
            //   "xterm"
            //   "konsole"
            //   "gnome-terminal"      // mustQuote = true
            //   "x-terminal-emulator" //alternative for konsole or gnome-terminal
        }


        Runnable processOutputMonitor = () -> {
            try {

                if (isMac) {
                    LOG.info("ExternalTerminal Thread: Running external terminal with command: " + finalCommand);
                    //cfr https://developer.apple.com/library/mac/documentation/Java/Conceptual/Java14Development/07-NativePlatformIntegration/NativePlatformIntegration.html
                    ScriptEngineManager mgr = new ScriptEngineManager();
                    ScriptEngine engine = mgr.getEngineByName("AppleScript");
                    if (engine != null) {
                        try {
                            Object result = engine.eval(finalCommand);

                            LOG.trace("ExternalTerminal Thread: Got output: " + result);
                        } catch (ScriptException e) {
                            LOG.error("ExternalTerminal Thread: Executing external ssh-terminal failed.", e);
                            throw new RuntimeException(e);
                        }
                    } else {
                        LOG.info("ExternalTerminal Thread: Got null when requesting AppleScript-engine, trying fallback to osascript");

                        ProcessBuilder pb = new ProcessBuilder("osascript", "-e", finalCommand);
                        try {
                            Process p = pb.start();

                            BufferedReader input = new BufferedReader(new InputStreamReader(p.getInputStream()));
                            BufferedReader input2 = new BufferedReader(new InputStreamReader(p.getErrorStream()));
                            String line;
                            while ((line = input.readLine()) != null) {
                                LOG.trace("ExternalTerminal Thread: External Terminal stdout: " + line);
                            }
                            while ((line = input2.readLine()) != null) {
                                LOG.trace("ExternalTerminal Thread: External Terminal stderr: " + line);
                            }
                            input.close();
                            input2.close();

                        } catch (IOException ex) {
                            LOG.error("ExternalTerminal Thread: Got exception while running ExternalTerminal-app", ex);
                            throw new RuntimeException(ex);
                        }

                    }
                } else {
                    //unix
                    LOG.debug("ExternalTerminal Thread: Starting terminal using command (list): " + launchTerminalCommandList);

                    ProcessBuilder pb = new ProcessBuilder(launchTerminalCommandList);
                    pb = pb.redirectErrorStream(true);

                    Process p = pb.start();
                    //                    Process p = Runtime.getRuntime().exec(finalCommand);
                    BufferedReader input = new BufferedReader(new InputStreamReader(p.getInputStream()));

                    String line;
                    while ((line = input.readLine()) != null) {
                        LOG.debug("ExternalTerminal Thread: External Terminal stdout+stderr: " + line);
                    }
                    //note: input might close before terminal stops.
                    input.close();
                    try {
                        p.waitFor();
                    } catch (InterruptedException e) {
                        LOG.warn("ExternalTerminal Thread: processOutputMonitor thread interrupted while waiting for terminal to stop", e);
                    }
                    LOG.info("ExternalTerminal Thread: Terminal command exited with exit value: " + p.exitValue());
                }
            } catch (Exception e) {
                LOG.error("Unhandled Exception in ExternalTerminal Thread.", e);
            } catch (AssertionError e) {
                LOG.error("Unhandled AssertionError in ExternalTerminal Thread.", e);
            }

            try {
                if (!jFedPreferences.getBoolean(HLPreferenceKey.PREF_TERMINAL_KEEP_TMP_FILES, true)) {
                    LOG.debug("ExternalTerminal Thread: Terminal has been shown. Will now wait 30s. (and then delete temp files)");
                    //some terminal commands exit immediately and start terminal in background
                    try {
                        Thread.sleep(30 * 1000);
                    } catch (InterruptedException e) {
                        //ignore
                    }

                    LOG.info("ExternalTerminal Thread: Terminal has been shown and waited 30s and will now delete temp files");
                    for (TmpFile tmpFile : tmpFiles)
                        tmpFile.delete();
                }
            } catch (AssertionError e) {
                LOG.error("Unhandled AssertionError in ExternalTerminal Thread.", e);
            } catch (Exception e) {
                LOG.error("Unhandled Exception in ExternalTerminal Thread.", e);
            }
        };
        try {
            for (TmpFile tmpFile : tmpFiles) {
                tmpFile.store();
            }

            Thread t = new Thread(processOutputMonitor);
            t.setName("ExternalTerminal");
            t.setDaemon(true); //=> this application may exit before terminal stops
            t.start();
        } catch (AssertionError e) {
            LOG.error("AssertionError starting ExternalTerminal Thread", e);
            if (!jFedPreferences.getBoolean(HLPreferenceKey.PREF_TERMINAL_KEEP_TMP_FILES, true)) {
                LOG.warn("  because of exception starting ExternalTerminal Thread -> will delete temp files");
                for (TmpFile tmpFile : tmpFiles)
                    tmpFile.delete();
            }
        } catch (Exception e) {
            LOG.error("Exception starting ExternalTerminal Thread", e);
            if (!jFedPreferences.getBoolean(HLPreferenceKey.PREF_TERMINAL_KEEP_TMP_FILES, true)) {
                LOG.warn("  because of exception starting ExternalTerminal Thread -> will delete temp files");
                for (TmpFile tmpFile : tmpFiles)
                    tmpFile.delete();
            }
        } catch (Error e) {
            //ok to catch, because thrown at end
            //noinspection finally
            try {
                LOG.error("Error starting ExternalTerminal Thread", e);
                if (!jFedPreferences.getBoolean(HLPreferenceKey.PREF_TERMINAL_KEEP_TMP_FILES, true)) {
                    LOG.warn("  because of exception starting ExternalTerminal Thread -> will delete temp files");
                    for (TmpFile tmpFile : tmpFiles)
                        tmpFile.delete();
                }
            } finally {
                throw e;
            }
        }

        if (isMac)
            try {
                //show a (mutable) notification to mac-users that the terminal opens in the background
                if (jFedPreferences.getBoolean(HLPreferenceKey.PREF_SHOW_MAC_TERMINAL_WARNING, true)) {
                    VBox pane = new VBox();
                    pane.setPrefWidth(750);
                    pane.setSpacing(5);

                    Label textLabel = new Label(String.format("An SSH-terminal to %s@%s:%d was launched.\n" +
                                    "It may however be on the background. Please switch applications to access your terminal.",
                            sshCommandBuilder.getTargetServerUserName(), hostName, sshCommandBuilder.getPort()));
                    textLabel.setPrefHeight(75);

                    textLabel.setWrapText(true);
                    pane.getChildren().add(textLabel);
                    final CheckBox checkBox = new CheckBox("Do not show this warning again");
                    pane.getChildren().add(checkBox);


                    DialogPane dialogPane = new DialogPane();
                    // bvermeul, FEDIBBTDEV-1049
                    dialogPane.setContent(pane);
                    dialogPane.getButtonTypes().setAll(ButtonType.OK);


                    Dialog<ButtonType> dialog = new Dialog<>();
                    dialog.setTitle("Terminal was launched");
                    dialog.setDialogPane(dialogPane);

                    dialog.showAndWait();

                    //dialogPane.setContent(pane);


                    //Alert alert = new Alert(Alert.AlertType.INFORMATION);
                    //alert.setTitle("Terminal was launched");
                    //alert.setDialogPane(dialogPane);
                    //alert.show();

                    jFedPreferences.setBoolean(HLPreferenceKey.PREF_SHOW_MAC_TERMINAL_WARNING,
                            !checkBox.isSelected());
                }
            } catch (Exception e) {
                LOG.error("Exception while showing MAC terminal dialog. (will be ignored)", e);
            } catch (AssertionError e) {
                LOG.error("AssertionError while showing MAC terminal dialog. (will be ignored)", e);
            }
    }


    private enum DoubleProxyHandling { IGNORE_AND_USE_JFED_PROXY, IGNORE_AND_USE_TESTBED_PROXY, DIALOG }

    @Nonnull
    private SshCommandBuilder prepareSshCommand(@Nonnull String targetServerUserName,
                                                @Nonnull String hostName,
                                                int port,
                                                @Nullable JFedConnection.ProxyInfo testbedProxyInfo,
                                                DoubleProxyHandling doubleProxyHandling) {
        SshCommandBuilder res = sshCommandBuilderFactory.create();
        res.setTargetServerUserName(targetServerUserName);
        res.setHostName(hostName);
        res.setPort(port);

        assert geniUserProvider.isUserLoggedIn();

        final GeniUser geniUser = geniUserProvider.getLoggedInGeniUser();

        //This will set the proxyInfo to what is needed, based on if proxy exists, preferences and connectivity data.
        JFedConnection.ProxyInfo jFedProxyInfo = geniUser.getUserAuthorityServer() == null ? null :
                proxyPreferencesManager.getSshProxySettings(geniUser.getUserAuthorityServer().getTestbed().getProxies(), geniUser);


        if (testbedProxyInfo != null && jFedProxyInfo != null) {
            switch (doubleProxyHandling) {
                case IGNORE_AND_USE_JFED_PROXY: {
                    LOG.warn("testbed proxy is specified '{}', however, it will be overridden by the jFed proxy", testbedProxyInfo);
                    res.setProxyInfo(jFedProxyInfo);
                    break;
                }
                case IGNORE_AND_USE_TESTBED_PROXY: {
                    LOG.warn("jFed Proxy is active, however, it will be overridden by a custom testbed proxy '{}'", testbedProxyInfo);
                    res.setProxyInfo(testbedProxyInfo);
                    break;
                }
                case DIALOG: {
                    LOG.warn("Proxy is active, however, so is the custom testbed proxy '{}'. " +
                            "Will inform user that only testbed proxy will be used.", testbedProxyInfo);


                    //show warning to user that we cannot do double proxy support

                    //show a (mutable) notification to users that they have two proxies enabled
                    if (jFedPreferences.getBoolean(HLPreferenceKey.PREF_SHOW_DOUBLE_PROXY_WARNING_DIALOG, true)) {
                        LOG.debug("Showing slivers initialization dialog");

                        VBox pane = new VBox();
                        pane.setPrefWidth(360);
                        pane.setSpacing(5);

                        Label textLabel = new Label(
                                "The testbed defined a proxy that has to be used to connect to the node. " +
                                        "You also have the jFed-proxy enabled for SSH-connections.\n\n" +
                                        "Your SSH-terminal-application only supports one proxy at a time. " +
                                        "We will therefore try to connect by only using the proxy defined by the testbed.\n\n" +
                                        "If you are experiencing connectivity problems because of this, please contact support.");
                        textLabel.setWrapText(true);
                        pane.getChildren().add(textLabel);
                        final CheckBox checkBox = new CheckBox("Do not show this dialog again");
                        pane.getChildren().add(checkBox);

                        DialogPane dialogPane = new DialogPane();
                        dialogPane.setHeaderText("2 Proxies detected");
                        dialogPane.setGraphic(GlyphUtils.createDialogGlyph(FontAwesome.Glyph.TERMINAL, Color.BLACK));
                        dialogPane.setContent(pane);
                        dialogPane.getButtonTypes().setAll(ButtonType.OK);


                        Dialog<ButtonType> dialog = new Dialog<>();
                        dialog.setTitle("Warning: 2 Proxies detected");
                        dialog.setDialogPane(dialogPane);

                        dialog.showAndWait();

                        res.setProxyInfo(testbedProxyInfo);

                        jFedPreferences.setBoolean(HLPreferenceKey.PREF_SHOW_DOUBLE_PROXY_WARNING_DIALOG,
                                !checkBox.isSelected());
                    } else {
                        LOG.debug("Skipping slivers initializing dialog because user muted it before");
                    }
                    break;
                }
            }
        } else {
            if (testbedProxyInfo != null) {
                res.setProxyInfo(testbedProxyInfo);
            }
            if (jFedProxyInfo != null) {
                res.setProxyInfo(jFedProxyInfo);
            }
        }
        res.setUseSshAgent(jFedPreferences.getBoolean(HLPreferenceKey.PREF_SSHAGENT_USE, true));
        return res;
    }
}

