package be.iminds.ilabt.jfed.experimenter_gui.editor.views;

import be.iminds.ilabt.jfed.experimenter_gui.canvas.rspec.CanvasNodeFactory;
import be.iminds.ilabt.jfed.experimenter_gui.config.JFedGuiConfig;
import be.iminds.ilabt.jfed.experimenter_gui.editor.ModelRspecEditor;
import be.iminds.ilabt.jfed.experimenter_gui.editor.ToolBoxFilter;
import be.iminds.ilabt.jfed.experimenter_gui.editor.ToolboxItem;
import be.iminds.ilabt.jfed.experimenter_gui.editor.addresspool_properties.AddressPoolPropertiesDialogFactory;
import be.iminds.ilabt.jfed.experimenter_gui.editor.channel_properties.ChannelPropertiesDialogFactory;
import be.iminds.ilabt.jfed.experimenter_gui.editor.impl.AddressPoolToolboxItem;
import be.iminds.ilabt.jfed.experimenter_gui.editor.impl.ChannelToolboxItem;
import be.iminds.ilabt.jfed.experimenter_gui.editor.impl.GeantTestbedToolboxItem;
import be.iminds.ilabt.jfed.experimenter_gui.editor.impl.NodeResourceClassToolboxItem;
import be.iminds.ilabt.jfed.experimenter_gui.util.ImageSize;
import be.iminds.ilabt.jfed.experimenter_gui.util.ImageUtil;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.ResourceClass;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Server;
import be.iminds.ilabt.jfed.lowlevel.testbed_info.TestbedInfoSource;
import be.iminds.ilabt.jfed.lowlevel.authority.finder.AuthorityFinder;
import be.iminds.ilabt.jfed.gui_preferences.HLPreferenceKey;
import be.iminds.ilabt.jfed.gui_preferences.JFedHLPreferences;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.google.common.eventbus.EventBus;
import javafx.application.Platform;
import javafx.beans.property.ObjectProperty;
import javafx.beans.property.SimpleObjectProperty;
import javafx.beans.property.SimpleStringProperty;
import javafx.beans.property.StringProperty;
import javafx.collections.ListChangeListener;
import javafx.event.EventHandler;
import javafx.geometry.Point2D;
import javafx.scene.Node;
import javafx.scene.control.Label;
import javafx.scene.control.ScrollPane;
import javafx.scene.input.*;
import javafx.scene.layout.*;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.Objects;
import java.util.function.Predicate;
import java.util.function.Supplier;
import java.util.stream.Collectors;

/**
 * User: twalcari
 * Date: 5/19/2015
 * Time: 16:39
 */
public class EditableCanvasEditorView extends BorderPane implements EditableRspecView {
    private static final Logger LOG = LoggerFactory.getLogger(EditableCanvasEditorView.class);

    private static final String EDITABLE_CANVAS_EDITOR_VIEW_FXML = "EditableCanvasEditorView.fxml";

    private final ModelRspecEditor modelRspecEditor;

    private final JFedGuiConfig conf;
    private final TestbedInfoSource testbedInfoSource;

    private final JFedHLPreferences preferences;

    private final EditableExperimentCanvas canvas;
    private final VBox toolbox;
    private final ToolBoxFilterPane toolBoxFilterPane;
    private final TilePane resourcesPane;

    private final StringProperty statusProperty = new SimpleStringProperty();

    private final ObjectProperty<ToolBoxFilter> toolBoxFilter = new SimpleObjectProperty<>(new ToolBoxFilter());

    private final List<ToolboxItem> toolboxItems = new ArrayList<>();

    private final EventHandler<MouseEvent> toolboxOnDragDetectedHandler = mouseEvent -> {
        //set selected toolboxitem as draggedItem
        ToolboxItem originalItem = (ToolboxItem) mouseEvent.getSource();

        Dragboard dragboard = originalItem.startDragAndDrop(TransferMode.COPY);
        ClipboardContent content = new ClipboardContent();
        content.put(ToolboxItem.getDragDataFormat(), originalItem.getId());
        dragboard.setContent(content);
        dragboard.setDragView(ImageUtil.imageOrFallback(originalItem.getDragImage(), ImageSize.SMALL));

        setStatus("Drop the item somewhere in the canvas.");

        mouseEvent.consume();
    };

    private final EventHandler<DragEvent> toolboxOnDragDoneHandler = event -> {

        if (event.getGestureTarget() == null)
            setStatus("Item can only be dropped on the canvas");

        event.consume();
    };


    EditableCanvasEditorView(ModelRspecEditor modelRspecEditor, JFedGuiConfig conf, CanvasNodeFactory canvasNodeFactory,
                             TestbedInfoSource testbedInfoSource, AuthorityFinder authorityFinder, ChannelPropertiesDialogFactory channelPropertiesDialogFactory,
                             AddressPoolPropertiesDialogFactory addressPoolPropertiesDialogFactory,
                             EventBus eventBus, EditToolBoxFiltersDialogFactory editToolBoxFiltersDialogFactory,
                             JFedHLPreferences preferences) {

        this.modelRspecEditor = modelRspecEditor;
        this.conf = conf;
        this.testbedInfoSource = testbedInfoSource;
        this.preferences = preferences;

        toolBoxFilter.setValue(getPreferencesToolBoxFilter());

        getStylesheets().add(getClass().getResource("ExperimenterEditor.css").toExternalForm());

        this.canvas =
                new EditableExperimentCanvas(modelRspecEditor, conf, canvasNodeFactory, testbedInfoSource, authorityFinder,
                        channelPropertiesDialogFactory, addressPoolPropertiesDialogFactory, eventBus);
        this.canvas.statusProperty().addListener(observable -> setStatus(canvas.getStatus()));

        this.canvas.setOnDragOver(event -> {
            if (event.getDragboard().hasContent(ToolboxItem.getDragDataFormat())) {
                event.acceptTransferModes(TransferMode.COPY);
            }
            event.consume();
        });

        this.canvas.setOnDragDropped(event -> {
            String draggedItemId =
                    (String) event.getDragboard().getContent(ToolboxItem.getDragDataFormat());
            ToolboxItem draggedItem = findToolboxItemById(draggedItemId);

            Point2D location = new Point2D(
                    event.getSceneX() - getCanvas().getLocalToSceneTransform().getTx(),
                    event.getSceneY() - getCanvas().getLocalToSceneTransform().getTy()
            );

            //perform this action out-of-band
            Platform.runLater(() -> draggedItem.onAddedToCanvas(canvas, location));

            event.consume();
        });

        setCenter(canvas);

        this.toolBoxFilterPane = new ToolBoxFilterPane(editToolBoxFiltersDialogFactory);
        toolBoxFilterPane.setFilter(toolBoxFilter.getValue());
        toolBoxFilterPane.setMaxWidth(210);

        this.resourcesPane = new TilePane();
        resourcesPane.getStyleClass().add("toolbox-pane");
        resourcesPane.setPrefWidth(200);

        ScrollPane resourcesScrollPane = new ScrollPane();
        resourcesScrollPane.setContent(resourcesPane);
        resourcesScrollPane.setPrefWidth(210);
        resourcesScrollPane.setHbarPolicy(ScrollPane.ScrollBarPolicy.NEVER);
        resourcesScrollPane.setVbarPolicy(ScrollPane.ScrollBarPolicy.AS_NEEDED);

        //TitledPane refused to behave well with a VBox in it instead of the ScrollPane.
        //It has been replaced by CSS as punishment
//        VBox toolBoxPane = new VBox();
//        toolBoxPane.setSpacing(0);
//        toolBoxPane.getChildren().addAll(toolBoxFilterPane, resourcesScrollPane);
//
//        TitledPane resourcesTitledPane = new TitledPane();
//        resourcesTitledPane.setText("Resources");
///       resourcesTitledPane.setContent(resourcesScrollPane);
//        resourcesTitledPane.setCollapsible(false);

        Label toolBoxTitle = new Label("Resources");
        toolBoxTitle.getStyleClass().add("toolbox-title");
        toolBoxTitle.setPrefWidth(210);
        VBox.setVgrow(toolBoxTitle, Priority.NEVER);

        this.toolbox = new VBox();
//        toolbox.getChildren().add(resourcesTitledPane);
        toolbox.getChildren().addAll(toolBoxTitle, toolBoxFilterPane, resourcesScrollPane);

        resourcesPane.getChildren().addListener((ListChangeListener<Node>) c -> {
            while (c.next()) {
                for (Node node : c.getAddedSubList()) {
                    assert (node instanceof ToolboxItem);
                    registerToolboxDragHandlers((ToolboxItem) node);

                }
            }
        });
        resourcesPane.prefHeightProperty().bind(toolbox.heightProperty());

        setLeft(toolbox);

        //for testing purposes:
//        toolBoxFilter.set(
//                new ToolBoxFilterBuilder()
////                        .addTestbedId("vwall2")
////                        .addTestbedId("vwall1")
//                        .addOrganisationId("iminds")
////                        .addCategoryName("cloud")
////                        .addFederationId("futebol")
//                        .create()
//                );

        //load toolbox content
        loadToolboxContent();

        //a bit of a hack with assert: listen and assert
        assert new Supplier<Boolean>() {
            @Override
            public Boolean get() {
                toolBoxFilter.addListener((observable, oldValue, newValue) -> { assert newValue != null; });
                return true;
            }
        }.get();

        toolBoxFilter.bind(toolBoxFilterPane.filterProperty());

        toolBoxFilter.addListener((observable, oldValue, newValue) -> loadToolboxContent());
        toolBoxFilter.addListener((observable, oldValue, newValue) -> setPreferencesToolBoxFilter(newValue));

        modelRspecEditor.setEditableExperimentCanvas(canvas);
    }

    private static final ObjectMapper MAPPER = new ObjectMapper();
    @Nonnull
    private ToolBoxFilter getPreferencesToolBoxFilter() {
        @Nullable String toolBoxfilterJson = preferences.getString(HLPreferenceKey.PREF_TOOLBOX_PREV_FILTER);
        if (toolBoxfilterJson == null)
            return new ToolBoxFilter();
        else {
            try {
                return MAPPER.readValue(toolBoxfilterJson, ToolBoxFilter.class);
            } catch (IOException e) {
                LOG.warn("Failed to restore ToolBoxFilter. Falling back to no filter.", e);
                return new ToolBoxFilter();
            }
        }
    }

    private void setPreferencesToolBoxFilter(@Nonnull ToolBoxFilter toolBoxFilter) {
        try {
            String toolBoxfilterJson = MAPPER.writeValueAsString(toolBoxFilter);
            preferences.setString(HLPreferenceKey.PREF_TOOLBOX_PREV_FILTER, toolBoxfilterJson);
        } catch (JsonProcessingException e) {
            LOG.warn("Failed to write ToolBoxFilter. not saving filter.", e);
        }
    }

    private ToolboxItem findToolboxItemById(@Nonnull String draggedItemId) {
        return toolboxItems.stream()
                .filter(item -> draggedItemId.equals(item.getId())).findFirst()
                .orElseThrow(() -> new RuntimeException("Could not find ToolboxItem with id " + draggedItemId));
    }

    private static final String WIRELESS_CHANNEL_ID_PREFIX = "wireless-channel";
    private static final String GEANT_TESTBED_ID_PREFIX = "geant-testbed";
    private static final String ADDRESS_POOL_ID_PREFIX = "address-pool";

    /**
     * Filters out special resource classes that have their own toolbox item
     */
    private static boolean isNormalResourceClass(ResourceClass resourceClass) {
        if (resourceClass.getId().startsWith(WIRELESS_CHANNEL_ID_PREFIX))
            return false;
        if (resourceClass.getId().startsWith(GEANT_TESTBED_ID_PREFIX))
            return false;
        if (resourceClass.getId().startsWith(ADDRESS_POOL_ID_PREFIX))
            return false;
        return true;
    }

    public void setToolBoxFilter(@Nonnull ToolBoxFilter toolBoxFilter) {
        if (toolBoxFilter == null)
            toolBoxFilter = new ToolBoxFilter();
        this.toolBoxFilter.set(toolBoxFilter);
    }
    @Nullable
    public ToolBoxFilter getToolBoxFilter() {
        return this.toolBoxFilter.get();
    }
    @Nonnull
    public ObjectProperty<ToolBoxFilter> toolBoxFilterProperty() {
        return toolBoxFilter;
    }

    private void loadToolboxContent() {
        assert toolBoxFilter.get() != null;
        Predicate<ResourceClass> resourceClassFilter = toolBoxFilter.get().getResourceClassFilter();
        Predicate<Server> serverFilter = toolBoxFilter.get().getServerFilter();

        toolboxItems.clear();
        toolboxItems.addAll(
                conf.getResourceClasses().stream()
                        .filter(item -> Objects.equals(item.getVisibleInToolbox(), Boolean.TRUE))
                        .filter(EditableCanvasEditorView::isNormalResourceClass)
                        .filter(resourceClassFilter)
                        .map(NodeResourceClassToolboxItem::new)
                        .collect(Collectors.toList()));

        if (toolBoxFilter.get().isEmpty() ||
                testbedInfoSource.getServers().stream()
                        .filter(server -> server.hasFlag(Server.Flag.featureWirelessChannelReservation))
                        .filter(serverFilter)
                        .findAny().isPresent()) {
            if (conf.getResourceClasses().stream()
                    .filter(item -> Objects.equals(item.getVisibleInToolbox(), Boolean.TRUE))
                    .filter(item -> item.getId().startsWith(WIRELESS_CHANNEL_ID_PREFIX))
                    .findAny().isPresent()) {
                toolboxItems.add(new ChannelToolboxItem());
            }
        }
        if (toolBoxFilter.get().isEmpty()){
            if (conf.getResourceClasses().stream()
                    .filter(item -> Objects.equals(item.getVisibleInToolbox(), Boolean.TRUE))
                    .filter(item -> item.getId().startsWith(GEANT_TESTBED_ID_PREFIX))
                    .findAny().isPresent()) {
                toolboxItems.add(new GeantTestbedToolboxItem());
            }
        }
        if (toolBoxFilter.get().isEmpty() ||
                testbedInfoSource.getServers().stream()
                        .filter(server -> server.hasFlag(Server.Flag.featureIPv4AddressPool))
                        .filter(serverFilter)
                        .findAny().isPresent()) {
            if (conf.getResourceClasses().stream()
                    .filter(item -> Objects.equals(item.getVisibleInToolbox(), Boolean.TRUE))
                    .filter(item -> item.getId().startsWith(ADDRESS_POOL_ID_PREFIX))
                    .findAny().isPresent()) {
                toolboxItems.add(new AddressPoolToolboxItem());
            }
        }

        toolboxItems.sort(Comparator.naturalOrder());

        resourcesPane.getChildren().setAll(toolboxItems);
    }


    private void registerToolboxDragHandlers(final ToolboxItem toolboxItem) {
        toolboxItem.setOnDragDetected(toolboxOnDragDetectedHandler);
        toolboxItem.setOnDragDone(toolboxOnDragDoneHandler);
    }

    public void setStatus(String status) {
        statusProperty.set(status);
    }

    public StringProperty statusProperty() {
        return statusProperty;
    }

    @Override
    public Region getView() {
        return this;
    }

    @Override
    public EditorViewType getType() {
        return EditorViewType.CANVAS;
    }

    @Override
    public RequestRspecSource getRequestRspecSource() {
        return new RequestRspecSource(modelRspecEditor.getModelRspec());
    }

    @Override
    public void dispose() {
        modelRspecEditor.setEditableExperimentCanvas(null);
    }

    public EditableExperimentCanvas getCanvas() {
        return canvas;
    }

    public ModelRspecEditor getModelRspecEditor() {
        return modelRspecEditor;
    }

    public void autoLayout() {
        canvas.autoLayout();
    }
    public void arrangeNodesOnCircle() {
        canvas.arrangeNodesOnCircle();
    }
    public void autoFit() {
        Platform.runLater(canvas::autoFit);
    }
    public void autoFit(double screenWidth, double screenHeight) {
        Platform.runLater(() -> canvas.autoFit(screenWidth, screenHeight));
    }
}
